from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.postgres.search import (
    SearchQuery,
    SearchRank,
    SearchVector,
    TrigramSimilarity,
)
from django.core.paginator import Paginator
from django.db import models
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse_lazy
from django.views.generic import CreateView, DeleteView, ListView, UpdateView

from .models import (
    Category,
    CategoryForm,
    Invoice,
    InvoiceForm,
    Snapshot,
    SnapshotForm,
    Transaction,
    TransactionForm,
)


@login_required
def index(request):
    _transactions = Transaction.objects.all()[:10]
    _categories = Category.objects.all()
    _snapshots = Snapshot.objects.all()

    context = {
        "transactions": _transactions,
        "categories": _categories,
        "snapshots": _snapshots,
    }
    return render(request, "main/index.html", context)


class LoginView(auth_views.LoginView):
    template_name = "main/login.html"
    next_page = "index"


class LogoutView(auth_views.LogoutView):
    next_page = "login"


class TransactionListView(LoginRequiredMixin, ListView):
    paginate_by = 24
    model = Transaction
    template_name = "main/transactions.html"
    context_object_name = "transactions"


class TransactionCreateView(LoginRequiredMixin, CreateView):
    model = Transaction
    form_class = TransactionForm


class InvoiceCreateView(LoginRequiredMixin, CreateView):
    model = Invoice
    form_class = InvoiceForm

    def form_valid(self, form):
        form.instance.transaction = get_object_or_404(
            Transaction, pk=self.kwargs["transaction_pk"]
        )
        return super().form_valid(form)

    def get_success_url(self):
        return reverse_lazy("transaction", kwargs={"pk": self.object.transaction.pk})


class CategoryCreateView(LoginRequiredMixin, CreateView):
    model = Category
    form_class = CategoryForm


class SnapshotCreateView(LoginRequiredMixin, CreateView):
    model = Snapshot
    form_class = SnapshotForm


class TransactionUpdateView(LoginRequiredMixin, UpdateView):
    model = Transaction
    form_class = TransactionForm


class InvoiceUpdateView(LoginRequiredMixin, UpdateView):
    model = Invoice
    form_class = InvoiceForm

    def get_success_url(self):
        return reverse_lazy("transaction", kwargs={"pk": self.object.transaction.pk})

    def get_queryset(self):
        return Invoice.objects.filter(
            transaction=get_object_or_404(Transaction, pk=self.kwargs["transaction_pk"])
        )


class CategoryUpdateView(LoginRequiredMixin, UpdateView):
    model = Category
    form_class = CategoryForm


class SnapshotUpdateView(LoginRequiredMixin, UpdateView):
    model = Snapshot
    form_class = SnapshotForm


class TransactionDeleteView(LoginRequiredMixin, DeleteView):
    model = Transaction
    template_name = "main/confirm_delete.html"
    success_url = reverse_lazy("index")


class InvoiceDeleteView(LoginRequiredMixin, DeleteView):
    model = Invoice
    template_name = "main/confirm_delete.html"

    def get_success_url(self):
        return reverse_lazy("transaction", kwargs={"pk": self.object.transaction.pk})

    def get_queryset(self):
        return Invoice.objects.filter(
            transaction=get_object_or_404(Transaction, pk=self.kwargs["transaction_pk"])
        )


class CategoryDeleteView(LoginRequiredMixin, DeleteView):
    model = Category
    template_name = "main/confirm_delete.html"
    success_url = reverse_lazy("index")


class SnapshotDeleteView(LoginRequiredMixin, DeleteView):
    model = Snapshot
    template_name = "main/confirm_delete.html"
    success_url = reverse_lazy("index")


@login_required
def search(request):
    _search = request.POST.get("search")
    return redirect("search", search=_search)


class SearchView(LoginRequiredMixin, ListView):
    paginate_by = 24
    model = Transaction
    template_name = "main/transactions.html"
    context_object_name = "transactions"

    def get_queryset(self):
        self.search = self.kwargs["search"]
        return (
            Transaction.objects.annotate(
                rank=SearchRank(
                    SearchVector("name", weight="A")
                    + SearchVector("description", weight="B")
                    + SearchVector("trader", weight="B"),
                    SearchQuery(self.search, search_type="websearch"),
                ),
                similarity=TrigramSimilarity("name", self.search),
            )
            .filter(models.Q(rank__gte=0.1) | models.Q(similarity__gte=0.3))
            .order_by("-rank", "-date")
        )

    def get_context_data(self, **kwargs):
        return super().get_context_data(**kwargs) | {"search": self.kwargs["search"]}
